#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <ctype.h>

#include "config.h"
#include "q2-wrapper.h"

struct directories dir;
struct settings config;


int	quit_verbosity;

void greetings(void)
/*
 * Spew some greetings
 *
 */
{
#ifndef _DEBUG_
	printf("%s version %s.\n", NAME, VERSION);
	printf("Copyright %s by %s <%s>\n", YEAR, AUTHOR, EMAIL);
	printf("Distributed under the GNU General Public Licence version 2.\n");
	printf("This software is NOT supported by Id Software or Activision.\n\n");
#endif
}

/************************************************************************/

void die(void)
{
#ifndef _DEBUG_
	if (quit_verbosity == BIG_QUIT)
		printf("\n");
		greetings();
#endif
	if (quit_verbosity > TERSE_QUIT) 
		printf("==== Q2-wrapper exiting ====\n\n");
}

/************************************************************************/

void usage(char *progname)
{
	printf("\nAll options are set in $HOME/%s/%s\n", 
			config.user_dir, config.user_config );

	printf("Read the q2-wrapper manpage for more information.\n");
	quit_verbosity = TERSE_QUIT;
	exit(0);
}

/************************************************************************/

void makeline(char lineout[LINELEN], int argc, char *argv[])
{
	int num;

	lineout[0] = 0;

	strcat(lineout, "+set basedir ");
	strncat(lineout, config.basedir, PATHLEN);

	strcat(lineout, "+set vid_ref ");
	strncat(lineout, config.vid_ref, PATHLEN);

	strcat(lineout, "+set cddir ");
	strncat(lineout, config.cddir, PATHLEN);

	strcat(lineout, "+set cd_dev ");
	strncat(lineout, config.cd_dev, PATHLEN);

	if (config.nocdaudio != 1) {
		strcat(lineout, "+set nocdaudio ");
		sprintf(lineout, "%d", config.nocdaudio);
	}

	if (config.sndbits != 0) {
		strcat(lineout, "+set sndbits ");
		sprintf(lineout, "%d", config.sndbits);
	}

	if (config.sndspeed != 0) {
		strcat(lineout, "+set sndspeed ");
		sprintf(lineout, "%d", config.sndspeed);
	}

	if (config.sndchannels != 2) {
		strcat(lineout, "+set sndchannels ");
		sprintf(lineout, "%d", config.sndchannels);
	}

	if (config.nostdout != 0) {
		strcat(lineout, "+set nostdout ");
		sprintf(lineout, "%d", config.nostdout);
	}

	
	for (num = 1; num <= argc; num++) {
		if (strlen(lineout) < LINELEN)
			strncat(lineout, argv[num], LINELEN);
		else {
			fprintf(stderr, "Too much stuff on the command line.\n");
			fprintf(stderr, "Extra dropped.\n");
		}
	}

}

/************************************************************************/

int getpaths(void)
{
	char 	*home;

	if ((home = getenv("HOME")) == NULL) {
		return(-1);	/* shouldn't ever happen */
	}

	dir.home[0]=0;
	dir.dot[0]=0;
	dir.save[0]=0;
	dir.pic[0]=0;
	dir.demo[0]=0;
	dir.players[0]=0;
	dir.video[0]=0;

	strncpy(dir.home, home, PATHLEN);

	/* Construct "/home/joe/.quake2" */
	strncpy(dir.dot, dir.home, PATHLEN);
	strcat(dir.dot, "/"); 
	strncat(dir.dot, config.user_config, PATHLEN); 

	/* seed the others with the dot directory */
	strncat(dir.save, dir.dot, PATHLEN);
	strncat(dir.pic, dir.dot, PATHLEN);
	strncat(dir.demo, dir.dot, PATHLEN);
	strncat(dir.players, dir.dot, PATHLEN);
	strncat(dir.video, dir.dot, PATHLEN);

	/* put the last part on */
	strcat(dir.save, "/"); 
	strncat(dir.save, config.savedir, PATHLEN);
	strcat(dir.pic, "/"); 
	strncat(dir.pic, config.picdir, PATHLEN);
	strcat(dir.demo, "/"); 
	strncat(dir.demo, config.demodir, PATHLEN);
	strcat(dir.players, "/"); 
	strncat(dir.players, config.playersdir, PATHLEN);
	strcat(dir.video, "/");
	strncat(dir.video, config.videodir, PATHLEN);

	/* the other exit shouldn't ever happen. */
	return(1);
}

/************************************************************************/

int checkdirs(void)
{
	char	skindir[PATHLEN];
	char	skins[PATHLEN];
	char	videodir[PATHLEN];
	

	/* For some odd reason, I need to #include <sys/stat.h> to make */
	/* gcc shut up about mkdir(2) being implicitly declared here. */

	if (chdir(dir.dot) != 0) {
		if (errno == ENOENT) {
			printf("%s does not exist, making...\n", dir.dot); 
			if (mkdir(dir.dot, DIRMODE) != 0) {
				perror("Problem making $HOME/.quake2\n");
				return(-1);
			}
			if (chown(dir.dot, getuid(), getgid()) != 0) {
				perror("Problem chowning $HOME/.quake2\n");
				return(-1);
			}			
		}
		else
			return(-1);
	}

	if (chdir(dir.save) != 0) {
		if (errno == ENOENT) {
			if (mkdir(dir.save, DIRMODE) != 0) {
				perror("Problem making the save directory.\n");
				return(-1);
			}
			if (chown(dir.save, getuid(), getgid()) != 0) {
				perror("Problem chowning the save directory.\n");
				return(-1);
			}
		}
		else
			return(-1);
	}

	if (chdir(dir.pic) != 0) {
		if (errno == ENOENT) {
			if (mkdir(dir.pic, DIRMODE) != 0) {
				perror("Problem making the srnshots directory.\n");
				return(-1);
			}
			if (chown(dir.pic, getuid(), getgid()) != 0) {
				perror("Problem chowning the scrnshots directory.\n");
				return(-1);
			}
		}
		else
			return(-1);
	}

	if (chdir(dir.demo) != 0) {
		if (errno == ENOENT) {
			if (mkdir(dir.demo, DIRMODE) != 0) {
				perror("Problem making the demo directory.\n");
				return(1);
			}
			if (chown(dir.demo, getuid(), getgid()) != 0) {
				perror("Problem chowning the demo directory.\n");
				return(-1);
			}
		}
		else
			return(-1);
	}

	if (chdir(dir.players) != 0) {
		if (errno == ENOENT) {
			if (mkdir(dir.players, DIRMODE) != 0) {
				perror("Problem making the players directory.\n");
				return(-1);
			}
			if (chown(dir.players, getuid(), getgid()) != 0) {
				perror("Problem chowning the players directory.\n");
				return(-1);
			}
		}
		else
			return(-1);
	}

	if (chdir(dir.video) != 0) {
		if (errno == ENOENT) {
			videodir[0] = 0;
			strncat(videodir, config.sys_baseq2, PATHLEN);
			strcat(videodir, "/");
			strncat(videodir, config.videodir, PATHLEN);
			if (symlink(videodir, dir.video) != 0) {
				perror("Problem symlinking to the video directory\n");
				return(-1);
			}
			if (chown(dir.video, getuid(), getgid()) != 0) {
				perror("Problem chowning the video directory symlink.\n");
				return(-1);
			}
		}
		else
			return(-1);
	}	

	skins[0] = 0;
	strncat(skins, dir.players, PATHLEN);
	strncat(skins, "/male", PATHLEN);
	if (chdir(skins) != 0) {
		if (errno == ENOENT) {
			skindir[0] = 0;
			strncat(skindir, config.sys_baseq2, PATHLEN);
			strcat(skindir, "/");
			strncat(skindir, config.playersdir, PATHLEN);
			strncat(skindir, "/male", PATHLEN);
			if(symlink(skindir, skins) != 0) {
				perror("Problem symlinking to the male skins directory\n");
				return(-1);
			}
			if(chown(skins, getuid(), getgid()) != 0) {
				perror("Problem chowning male skins symlink.\n");
				return(-1);
			}
		}
		else
			return(-1);
	}

	skins[0] = 0;
	strncat(skins, dir.players, PATHLEN);
	strncat(skins, "/female", PATHLEN);
	if (chdir(skins) != 0) {
		if (errno == ENOENT) {
			skindir[0] = 0;
			strncat(skindir, config.sys_baseq2, PATHLEN);
			strcat(skindir, "/");
			strncat(skindir, config.playersdir, PATHLEN);
			strncat(skindir, "/female", PATHLEN);
			if(symlink(skindir, skins) != 0) {
				perror("Problem symlinking to the female skins directory\n");
				return(-1);
			}
			if(chown(skins, getuid(), getgid()) != 0) {
				perror("Problem chowning female skins symlink.\n");
				return(-1);
			}
		}
		else
			return(-1);
	}

	return(1);
}

/************************************************************************/

int makelinks(void)
{
	char todir[PATHLEN+1];


	todir[0] = 0;
	strncat(todir, config.sys_baseq2, PATHLEN);

	if (symlink(dir.dot, todir) != 0) {
		perror("Problem symlinking to $HOME/.quake2\n");
		return(-1);
	}
	return(1);
}

/************************************************************************/

int breaklinks(void)
{
	char todir[PATHLEN+1];

	todir[0] = 0;
	strncat(todir, config.sys_baseq2, PATHLEN);

	if (unlink(todir) != 0) {
		perror("Problem removing symlink to $HOME/.quake2\n");
		return(-1);
	}
	return(1);
}

/************************************************************************/

int initconfigvars(struct settings *var)
{
	var->basedir[0] = 0;	
	var->baseq2[0] = 0;	
	var->binpath[0] = 0;	
	var->sys_baseq2[0] = 0;	
	var->cddir[0] = 0;	
	var->cd_dev[0] = 0;	
	var->user_config[0] = 0;	
	var->user_dir[0] = 0;	
	var->vid_ref[0] = 0;	

	var->savedir[0] = 0;	
	var->picdir[0] = 0;	
	var->demodir[0] = 0;	
	var->playersdir[0] = 0;	
	var->sys_playersdir[0] = 0;	

	var->nocdaudio = 1;
	var->sndbits = 0;
	var->sndspeed = 0;
	var->sndchannels = 2;
	var->nostdout = 0;

/*
	var->dirmode = 000;
	var->filemode = 000;
*/
	return(1);

}

/************************************************************************/

int getconfig(int sys_or_user)
{
	FILE	*fp;

	int	num;
	int	num2;

	int	linecount;

	char	varname[LINELEN + 1];
	char	value[LINELEN + 1];
	char	thisconfig[LINELEN + 1];


	if (sys_or_user == SYS_CONFIG_CHOICE) {
		if ((fp = fopen(SYS_CONFIG, "r")) == NULL) {
			fprintf(stderr, "Problem opening %s\n", SYS_CONFIG);
			perror("This is rather important\n");
			fclose(fp);
			return(-1);
		}
		strncpy(thisconfig, SYS_CONFIG, LINELEN);
		initconfigvars(&config);
	}
	else if (sys_or_user == USER_CONFIG_CHOICE) {
		if ((fp = fopen(config.user_config, "r")) == NULL) 
			perror("Can't open user's q2-wrapper config file\n");
		strncpy(thisconfig, config.user_config, LINELEN);
		/* so what?  we'll just use the system settings */
	}
	else {
		fprintf(stderr, "Fatal error calling getconfig.  Expected SYS_CONFIG or USER_CONFIG\n");
		exit(1);
	}


	initconfigvars(&config);

	linecount = 1;

	while (fgets(varname, LINELEN, fp) != NULL) {

		if (varname[strlen(varname)-1] != '\n')
			while (fgetc(fp) != '\n')
			;
		/* If we didn't get a whole line, dump the rest of the line */

		for (num = strlen(varname) - 1; isspace(varname[num]); num--)
		;
		varname[num+1] = 0;
		/* Removed trailing whitespace and newline */

		for (num = 0; num <= strlen(varname)+1; num++) {
			if (varname[num] == COMMENT1)
				varname[num] = 0;
			else if (strncmp(&varname[num], COMMENT2, COMMENTLEN) == 0)
				varname[num] = 0;
		}
		/* Detect "til end of line" comments */

		for (num = 0; !isspace(varname[num]) && num < NAMELEN; num++);
		/* got the end of the variable name */

		for (num2 = num; isspace(varname[num2]) && num2 < NAMELEN; num2++);
		/* for the beginning of the value */

		strncpy(value, &varname[num2], LINELEN);
		varname[num] = 0; /* chop off the value from the var name */

		/* varname now contains the variable name */


		/* we don't want lusers to be mucking with these settings */
		if (sys_or_user == SYS_CONFIG_CHOICE) {
			if (strncmp(varname, "basedir", PATHLEN) == 0)
				strncpy(config.basedir, value, PATHLEN);

			else if (strncmp(varname, "baseq2", PATHLEN) == 0)
				strncpy(config.baseq2, value, PATHLEN);

			else if (strncmp(varname, "binpath", PATHLEN) == 0)
				strncpy(config.binpath, value, PATHLEN);

			else if (strncmp(varname, "cddir", PATHLEN) == 0)
				strncpy(config.cddir, value, PATHLEN);

			else if (strncmp(varname, "cd_dev", PATHLEN) == 0)
				strncpy(config.cd_dev, value, PATHLEN);

			else if (strncmp(varname, "user_config", PATHLEN) == 0)
				strncpy(config.user_dir, value, PATHLEN);

			else if (strncmp(varname, "user_dir", PATHLEN) == 0)
				strncpy(config.user_dir, value, PATHLEN);

			else if (strncmp(varname, "savedir", PATHLEN) == 0)
				strncpy(config.savedir, value, PATHLEN);

			else if (strncmp(varname, "picdir", PATHLEN) == 0)
				strncpy(config.picdir, value, PATHLEN);

			else if (strncmp(varname, "demodir", PATHLEN) == 0)
				strncpy(config.demodir, value, PATHLEN);

			else if (strncmp(varname, "playersdir", PATHLEN) == 0)
				strncpy(config.playersdir, value, PATHLEN);

			else if (strncmp(varname, "videodir", PATHLEN) == 0)
				strncpy(config.videodir, value, PATHLEN);



			else if (strncmp(varname, "sys_baseq2", PATHLEN) == 0)
				strncpy(config.sys_baseq2, value, PATHLEN);

			else if (strncmp(varname, "sys_playersdir", PATHLEN) == 0)
				strncpy(config.sys_playersdir, value, PATHLEN);

			/* dirmode and filemode unimplemented */
		}


		/* Values set here will override system settings */


		if (strncmp(varname, "vid_ref", PATHLEN) == 0)
			strncpy(config.vid_ref, value, PATHLEN);


		else if (strncmp(varname, "nocdaudio", PATHLEN) == 0) {
			for (num = 0; value[num] != 0; num++)
				value[num] = tolower(value[num]);
			if (strncmp(value, "true", 4) == 0)
				config.nocdaudio = 1;  /* true */
			else if (strncmp(value, "false", 5) == 0)
				config.nocdaudio = 0;  /* false */
			else {
				fprintf(stderr, "Invalid value for \"nocdaudio\" on line %d in %s", linecount, thisconfig);
				fprintf(stderr, "Defaulting to TRUE\n");
				config.nocdaudio = 1;
			}
		}

		else if (strncmp(varname, "sndbits", PATHLEN) == 0) {
			for (num = 0; value[num] != 0; num++) {
				if (!isdigit(value[num])) {
					fprintf(stderr, "Whoops!  Nondigits in the \"sndbits\" field on line %d in %s\n", linecount, thisconfig);
					return(1);
				}
			}
			num = atoi(value);
			if (num <= 0 || num > 64) {
				fprintf(stderr, "\"sndbits\" value is unrealistic.  Line %d in %s.  Not set.\n", linecount, thisconfig);
				config.sndbits = 0;
			}
			config.sndbits = num;
		}

		else if (strncmp(varname, "sndspeed", PATHLEN) == 0) {
			for (num = 0; value[num] != 0; num++) {
				if (!isdigit(value[num])) {
					fprintf(stderr, "Whoops!  Nondigits in the \"sndspeed\" field on line %d in %s\n", linecount, thisconfig);
					return(1);
				}
			}
			num = atoi(value);
			if (num < 0 || num > 44100) {
				fprintf(stderr, "\"sndspeed\" value is unrealistic.  Line %d in %s.  Not set.\n", linecount, thisconfig);
				config.sndspeed = 0;
			}
			config.sndspeed = num;
		}

		else if (strncmp(varname, "sndchannels", PATHLEN) == 0) {
			for (num = 0; value[num] != 0; num++) {
				if (!isdigit(value[num])) {
					fprintf(stderr, "Whoops!  Nondigits in the \"sndchannels\" field on line %d in %s\n", linecount, thisconfig);
					return(1);
				}
			}
			num = atoi(value);
			if (num < 1 || num > 2) {
				fprintf(stderr, "\"sndspeed\" value is unrealistic.  Line %d in %s.  Not set.\n", linecount, thisconfig);
				config.sndchannels = 2;
			}
			config.sndchannels = num;
		}

		else if (strncmp(varname, "nostdout", PATHLEN) == 0) {
			for (num = 0; value[num] != 0; num++)
				value[num] = tolower(value[num]);
			if (strncmp(value, "true", 4) == 0)
				config.nostdout = 1;  /* true */
			else if (strncmp(value, "false", 5) == 0)
				config.nostdout = 0;  /* false */
			else {
				fprintf(stderr, "Invalid value for \"nostdout\" on line %d in %s", linecount, thisconfig);
				fprintf(stderr, "Defaulting to FALSE\n");
				config.nostdout = 0;
			}
		}

		else
			fprintf(stderr, "Unrecognized variable name on line %d in %s\n", linecount, thisconfig);

		linecount++;
	}

	return(1);
}
